
from typing import TYPE_CHECKING, List
    
if TYPE_CHECKING:
    import django.contrib.auth.models as dmodels

class Permission:
    name:     str
    codename: str

    def __init__(self, name: str, codename: str):
        self.name     = name
        self.codename = codename
    def __eq__(self, value: object) -> bool:
        return isinstance(value, Permission) \
           and self.name     == value.name \
           and self.codename == value.codename

    @staticmethod
    def from_permission (permission: "dmodels.Permission") -> "Permission":
        return Permission( permission.name, permission.codename )

class Group:
    name: str

    permissions: List[Permission]

    def __init__(self, name: str, permissions: List[Permission]):
        self.name = name

        self.permissions = permissions
    def __eq__(self, value: object) -> bool:
        return isinstance(value, Group) \
           and self.name == value.name \
           and all([ x in self.permissions for x in value.permissions ]) \
           and len(self.permissions) == len(value.permissions)

    @staticmethod
    def from_group (group: "dmodels.Group") -> "Group":
        dPermissions = group.permissions.all()

        permissions = [ Permission.from_permission(permission) for permission in dPermissions ]

        return Group(group.name, permissions)

class User:
    username: str

    is_anonymous     : bool
    is_authenticated : bool

    is_staff     : bool
    is_active    : bool
    is_superuser : bool

    groups: List[Group]

    def __init__(
            self,

            username : str,
            
            is_anonymous     : bool, 
            is_authenticated : bool,

            is_staff     : bool,
            is_active    : bool,
            is_superuser : bool,

            groups : List[Group]
        ) -> None:
        
        self.username = username

        self.is_anonymous     = is_anonymous
        self.is_authenticated = is_authenticated

        self.is_staff     = is_staff
        self.is_active    = is_active
        self.is_superuser = is_superuser

        self.groups = groups

    def __eq__(self, value: object) -> bool:
        return isinstance(value, User) \
           and self.username == value.username \
           and self.is_anonymous == value.is_anonymous \
           and self.is_authenticated == value.is_authenticated \
           and self.is_staff == value.is_staff \
           and self.is_active == value.is_active \
           and self.is_superuser == value.is_superuser \
           and len(self.groups) == len(value.groups) \
           and all([ x in value.groups for x in self.groups ])

    @staticmethod
    def from_user (user: "dmodels.User") -> "User":
        dGroups = user.groups.all()

        groups = [ Group.from_group(group) for group in dGroups ]

        return User(
            user.username,

            user.is_anonymous,
            user.is_authenticated,
            user.is_staff,
            user.is_active,
            user.is_superuser,

            groups
        )



from typing import Callable, Generic, TypeVar

from django.http import HttpRequest, JsonResponse
from django.urls import path
from statelessauth.engine.abstract import AuthEngine
from statelessauth.prometheus import engine_view_decorator, acquire_engine_acquire_metrics
from statelessauth.wire import AuthWire

T = TypeVar("T")

class AcquireEngine(Generic[T], AuthEngine[T]):
    __acquire_view: "Callable[[HttpRequest], T | None]"

    def __init__(
            self, 
            keyname: str,
            scheme: AuthWire[T],
            acquire_view: "Callable[[HttpRequest], T | None]",
            *args, **kwargs
            ):
        super().__init__(keyname, scheme, *args, **kwargs)

        self.__acquire_view = acquire_view

    @engine_view_decorator(acquire_engine_acquire_metrics)
    def acquire_view (self, request: HttpRequest, *args, **kwargs):
        result = self.__acquire_view(request, *args, **kwargs)
        if result is None:
            return JsonResponse({
                "valid": False,
                "token": ""
            }, status=401)
        
        return JsonResponse({
            "valid": True,
            "token": self.encode(result)
        }, status=200)

    @property
    def urlpatterns (self):
        return [
            path('acquire/', lambda *args, **kwargs: self.acquire_view(*args, **kwargs))
        ]




import time
import json
from typing import TYPE_CHECKING, Any, Callable, List, Type, TypeVar
from django.conf import settings
from django.http import JsonResponse
from prometheus_client      import REGISTRY, Counter, Histogram
from prometheus_client.metrics import MetricWrapperBase

stateless_auth_metrics_generated = False

class EngineMetric:
    __metric_fnc : "Callable[[Any]]"

    def __init__(self, metric_cls: Type[MetricWrapperBase], metric_fnc, name: str, help: str, labels: List[str] = [], *args, **kwargs):
        from django_prometheus.conf import NAMESPACE
        self.name = name
        self.help = help

        self.__metric_fnc = metric_fnc

        self.metric_total = metric_cls(
            name,
            help,
            *args, 
            namespace = NAMESPACE,
            **({ "labelnames" : labels, **kwargs } if len(labels) != 0 else kwargs)
        )
        self.metric_per_engine = metric_cls(
            name + "_per_engine",
            help + " per authentication engine",
            [ 'engine' ] + labels,
            *args, 
            namespace = NAMESPACE,
            **kwargs
        )
    def run_on (self, engine: str, ob_args: List = [],  lbl_args: List = []):
        self.__metric_fnc(
            self.metric_total.labels(*lbl_args) if len(lbl_args) != 0 else self.metric_total,
            *ob_args
        )
        self.__metric_fnc(
            self.metric_per_engine.labels(engine, *lbl_args),
            *ob_args
        )
class TotalAndLatencyMetric:
    def __init__(self, name: str, help1: str, help2: str):
        from django_prometheus.conf import PROMETHEUS_LATENCY_BUCKETS, NAMESPACE
        self.total = EngineMetric(
            Counter,
            Counter.inc,
            name + "_total",
            help1
        )
        self.latency = EngineMetric(
            Histogram,
            Histogram.observe,
            name + "_latency",
            help2,
            buckets=PROMETHEUS_LATENCY_BUCKETS
        )
    def run (self, name: str, latency: float):
        self.total.run_on(name)
        self.latency.run_on(name, [ latency ])

class MiddlewareMetrics:
    _was_init = False
    def _init (self):
        if self._was_init: return
        self._was_init = True
        from django_prometheus.conf import PROMETHEUS_LATENCY_BUCKETS, NAMESPACE

        self.stateless_auth_middleware_run = Counter(
            "stateless_auth_middleware_run",
            "Number of runs of the middleware",
            namespace=NAMESPACE
        )
        self.stateless_auth_middleware_run_latency = Histogram(
            "stateless_auth_middleware_run_latency",
            "Latency of the stateless auth middleware",
            buckets=PROMETHEUS_LATENCY_BUCKETS,
            namespace=NAMESPACE
        )

        self.stateless_auth_middleware_run_per_config = Counter(
            "stateless_auth_middleware_run_per_config",
            "Number of runs per sub middleware configured",
            [ "middleware" ],
            namespace=NAMESPACE
        )
        self.stateless_auth_middleware_run_per_config_latency = Histogram(
            "stateless_auth_middleware_run_latency_per_config",
            "Latency of the stateless auth middleware",
            [ "middleware" ],
            buckets=PROMETHEUS_LATENCY_BUCKETS,
            namespace=NAMESPACE
        )
        self.stateless_auth_middleware_run_per_config_missing = Counter(
            "stateless_auth_middleware_run_per_config_missing",
            "Number of runs per sub middleware configured where header is missing",
            [ "middleware" ],
            namespace=NAMESPACE
        )
        self.stateless_auth_middleware_run_per_config_wrong_token = Counter(
            "stateless_auth_middleware_run_per_config_wrong_token",
            "Number of runs per sub middleware configured where the token is invalid",
            [ "middleware" ],
            namespace=NAMESPACE
        )
        self.stateless_auth_middleware_run_per_config_no_type = Counter(
            "stateless_auth_middleware_run_per_config_no_type",
            "Number of runs per sub middleware configured where header has no type",
            [ "middleware" ],
            namespace=NAMESPACE
        )
        self.stateless_auth_middleware_run_per_config_valid = Counter(
            "stateless_auth_middleware_run_per_config_valid",
            "Number of runs per sub middleware configured where header is valid",
            [ "middleware" ],
            namespace=NAMESPACE
        )
    def clear (self):
        self._was_init = False
        self._init()

middleware_metrics = MiddlewareMetrics()

class TokenBasedViewMetrics:
    _was_init = False
    def __init__(self, name: str, qname: str):
        self.name = name
        self.qname = qname
    def _init(self):
        if self._was_init: return
        self._was_init = True
        from django_prometheus.conf import PROMETHEUS_LATENCY_BUCKETS, NAMESPACE
        self.total_counter = TotalAndLatencyMetric(
            self.qname,
            f"Total number of calls to {self.name}",
            f"Latency of calls to {self.name}"
        )
        self.success_counter = TotalAndLatencyMetric(
            self.qname + "_success",
            f"Total number of successful calls to {self.name}",
            f"Latency of successful calls to {self.name}"
        )
        self.failed_counter = TotalAndLatencyMetric(
            self.qname + "_failed",
            f"Total number of failed calls to {self.name}",
            f"Latency of failed calls to {self.name}"
        )
    def observe (self, engine: str, latency: float, valid: bool):
        self.total_counter.run( engine, latency )
        
        if valid:
            self.success_counter.run( engine, latency )
        else:
            self.failed_counter.run( engine, latency )

    def clear (self):
        self._was_init = False
        self._init()

acquire_engine_acquire_metrics = TokenBasedViewMetrics(
    "AcquireEngine.acquire_view",
    "stateless_auth_acquire_engine_acquire_view"
)
refresh_engine_acquire_metrics = TokenBasedViewMetrics(
    "RefreshEngine.acquire_view",
    "stateless_auth_refresh_engine_acquire_view"
)

class RefreshViewMetrics:
    _was_init = False
    def __init__(self, name: str, qname: str):
        self.name = name
        self.qname = qname
    def _init(self):
        if self._was_init: return
        self._was_init = True
        from django_prometheus.conf import PROMETHEUS_LATENCY_BUCKETS, NAMESPACE
        self.total_counter = TotalAndLatencyMetric(
            self.qname,
            f"Total number of calls to {self.name}",
            f"Latency of calls to {self.name}"
        )
        self.success_counter = TotalAndLatencyMetric(
            self.qname + "_success",
            f"Total number of successful calls to {self.name}",
            f"Latency of successful calls to {self.name}"
        )
        self.wrong_counter = TotalAndLatencyMetric(
            self.qname + "_wrong",
            f"Total number of calls with wrong tokens to {self.name}",
            f"Latency of calls with wrong tokens to {self.name}"
        )
        self.expired_counter = TotalAndLatencyMetric(
            self.qname + "_expired",
            f"Total number of calls with expired tokens to {self.name}",
            f"Latency of calls with expired tokens to {self.name}"
        )
        self.missing_counter = TotalAndLatencyMetric(
            self.qname + "_missing",
            f"Total number of calls where refresh token is missing to {self.name}",
            f"Latency of calls where refresh token is missing to {self.name}"
        )
    def clear (self):
        self._was_init = False
        self._init()

refresh_engine_refresh_metrics = RefreshViewMetrics(
    "RefreshEngine.refresh_view",
    "stateless_auth_refresh_engine_refresh_view"
)

class RunningOnMetric:
    start: int
    m1   : Counter
    m2   : Histogram
    def __init__(self, m1: Counter, m2: Histogram, labels: List[str] = []) -> None:
        if len(labels) != 0:
            m1 = m1.labels( *labels )
            m2 = m2.labels( *labels )
        self.m1 = m1
        self.m2 = m2

    def __enter__(self):
        self.m1.inc()
        self.start = time.time()
        return self
    def __exit__(self, *args):
        self.m2.observe( time.time() - self.start )

def generate_metrics ():
    global stateless_auth_encode

    global stateless_auth_decode
    global stateless_auth_decode_failed
    global stateless_auth_decode_success

    global stateless_auth_metrics_generated
    if stateless_auth_metrics_generated: return
    stateless_auth_metrics_generated = True

    stateless_auth_encode = TotalAndLatencyMetric(
        "stateless_auth_encode",
        "Total number of authentication encodes",
        "Latency during encoding of authentication data"
    )
    stateless_auth_decode = TotalAndLatencyMetric(
        "stateless_auth_decode",
        "Total number of authentication decodes",
        "Latency during encoding of authentication data"
    )
    stateless_auth_decode_success = TotalAndLatencyMetric(
        "stateless_auth_decode_success",
        "Total number of successful authentication decodes",
        "Latency during encoding of valid authentication data"
    )
    stateless_auth_decode_failed  = TotalAndLatencyMetric(
        "stateless_auth_decode_failed",
        "Total number of failed authentication decodes",
        "Latency during decoding of wrong authentication data"
    )

def uses_metrics (f):
    def wrapped (*args, **kwargs):
        generate_metrics()
        return f(*args, **kwargs)
    wrapped.__name__ = f.__name__
    return wrapped

def stateless_auth_encode_decorator (f):
    @uses_metrics
    def wrapped (self, *args, **kwargs):
        start = time.time()
        result = f(self, *args, **kwargs)
        end = time.time()

        delta = end - start
        
        stateless_auth_encode.run( self.name, delta )

        return result
    wrapped.__name__ = f.__name__
    return wrapped
def stateless_auth_decode_decorator (f):
    @uses_metrics
    def wrapped (self, *args, **kwargs):
        start = time.time()

        result = None
        exception = None
        try:
            result = f(self, *args, **kwargs)
        except Exception as exc:
            exception = exc
        end = time.time()

        delta = end - start
        
        stateless_auth_decode.run( self.name, delta )

        if exception is not None or result is None:
            if exception is not None: raise exception

            stateless_auth_decode_failed.run(self.name, delta)
        else:
            stateless_auth_decode_success.run(self.name, delta)

        return result
    wrapped.__name__ = f.__name__
    return wrapped
def middleware_decorator (f):
    def wrapped (self, *args, **kwargs):
        middleware_metrics._init()
        with RunningOnMetric(
            middleware_metrics.stateless_auth_middleware_run,
            middleware_metrics.stateless_auth_middleware_run_latency
        ):
            return f(self, *args, **kwargs)
    wrapped.__name__ = f.__name__
    return wrapped

def engine_view_decorator (engine_metrics: TokenBasedViewMetrics):
    def decorator (f):
        def wrapped (self, *args, **kwargs):
            start = time.time()
            result: JsonResponse = f(self, *args, **kwargs)
            end = time.time()

            engine_metrics._init()
            content = result.content
            data = json.loads( content.decode() )
            engine_metrics.observe( self.name, end - start, data["valid"] )
            return result
        wrapped.__name__ = f.__name__
        return wrapped
    return decorator

def clear_metrics ():
    collectors = list( REGISTRY._collector_to_names.keys() )
    for collector in collectors:
        REGISTRY.unregister(collector)
    global stateless_auth_metrics_generated
    stateless_auth_metrics_generated = False

    generate_metrics()
    
    middleware_metrics.clear()
    acquire_engine_acquire_metrics.clear()
    refresh_engine_acquire_metrics.clear()
    refresh_engine_refresh_metrics.clear()




from typing import TYPE_CHECKING, Dict, List, Self, Tuple, Union

from statelessauth.wire import AuthWire

if TYPE_CHECKING:
    from statelessauth.engine.abstract import AuthEngine

from jose.jws import Key
from django.conf import settings

class SAuthMiddlewareConfig:
    engine: str
    header: Tuple[str, str]
    field : str

    urls: List[str] | None

    def get_engine (self) -> "AuthEngine":
        return StatelessAuthConfig.instance().get_engine(self.engine)

    def __init__(self, obj: Dict = {}) -> None:
        self.engine = 'default'
        self.header = ('Authorization', 'Bearer')
        self.field  = 'unknown'
        self.urls   = None

        if 'engine' in obj:
            self.engine = obj['engine']
        if 'header' in obj:
            self.header = obj['header']
        if 'urls'   in obj:
            self.urls   = obj['urls']
        if 'field'  in obj:
            self.field  = obj['field']
    @staticmethod
    def create (input):
        if isinstance(input, SAuthMiddlewareConfig): return input
        return SAuthMiddlewareConfig(input)

class StatelessAuthConfig:
    SL_AUTH_KEY_BACKENDS = "SL_AUTH_KEY_BACKENDS"
    SL_AUTH_MIDDLEWARES  = "SL_AUTH_MIDDLEWARES"
    SL_AUTH_ENGINES      = "SL_AUTH_ENGINES"

    __config: "StatelessAuthConfig | None" = None

    __key_backends : Dict[str, Key]
    __middlewares  : List[Tuple[str, SAuthMiddlewareConfig]]
    __engines      : "Dict[str, AuthEngine]"

    @property
    def middlewares (self) -> List[Tuple[str, SAuthMiddlewareConfig]]:
        return self.__middlewares

    def get_key (self, name: str) -> "Key | None":
        return self.__key_backends.get(name, None)
    def get_engine (self, name: str) -> "AuthEngine | None":
        return self.__engines.get(name, None)
    def load_config (self):
        if hasattr(settings, self.SL_AUTH_KEY_BACKENDS):
            self.__key_backends = getattr(settings, self.SL_AUTH_KEY_BACKENDS)
        
        if hasattr(settings, self.SL_AUTH_MIDDLEWARES):
            middlewares: Dict = getattr(settings, self.SL_AUTH_MIDDLEWARES)
            self.__middlewares = [
                (key, SAuthMiddlewareConfig.create( middlewares[key] ))
                for key in middlewares.keys()
            ]
            self.__middlewares.sort(key = lambda key: key[0])

        if hasattr(settings, self.SL_AUTH_ENGINES):
            self.__engines = getattr(settings, self.SL_AUTH_ENGINES)

            for name in self.__engines.keys():
                engine = self.__engines[name]

                engine.name = name

    def __init__(self, load_config = True) -> None:
        if load_config:
            self.load_config()

    def __new__(cls, load_config = True) -> Self:
        if cls != StatelessAuthConfig or not load_config:
            return super().__new__(cls)
        
        if StatelessAuthConfig.__config is not None:
            return StatelessAuthConfig.__config

        StatelessAuthConfig.__config = super().__new__(cls)
        return StatelessAuthConfig.__config
    @staticmethod
    def instance():
        return StatelessAuthConfig()
    @staticmethod
    def create_config (key_backends, engines, middlewares):
        conf = StatelessAuthConfig(False)
        conf.__key_backends = key_backends
        conf.__engines      = engines
        conf.__middlewares  = [
            (key, SAuthMiddlewareConfig.create(middlewares[key]))
            for key in middlewares.keys()
        ]
        conf.__middlewares.sort(key = lambda key: key[0])

        return conf



import math
from typing import Callable, Generic, TypeVar

from django.http import HttpRequest, JsonResponse
from django.urls import path
from statelessauth.engine.abstract import AuthEngine
from statelessauth.prometheus import engine_view_decorator, refresh_engine_acquire_metrics, refresh_engine_refresh_metrics
from statelessauth.wire import AuthWire

import time

T = TypeVar("T")

class RefreshEngine(Generic[T], AuthEngine[T]):
    __acquire_view: "Callable[[HttpRequest], T | None]"

    __short_timeout: int
    __long_timeout : int

    __scheme : AuthWire[T]

    def __init__(
            self, 
            keyname: str,
            scheme: AuthWire[T],
            acquire_view: "Callable[[HttpRequest], T | None]",
            short_timeout = 5 * 60, # 5 minutes short timeout
            long_timeout  = 14 * 24 * 60 * 60, # 14 days long timeout
            *args, **kwargs
            ):
        super().__init__(keyname, scheme, *args, **kwargs)

        self.__acquire_view = acquire_view

        self.__scheme = scheme

        self.__short_timeout = short_timeout
        self.__long_timeout  = long_timeout

    def payload_from_wired(self, wired):
        start = time.time_ns()

        return {
            "wired": wired,
            "alt": start + math.floor(1_000_000_000 * self.__short_timeout),
            "rlt": start + math.floor(1_000_000_000 * self.__long_timeout)
        }
    def wired_from_payload(self, payload):
        return payload["wired"]
    def validate_payload(self, payload):
        return time.time_ns() <= payload['alt']

    @engine_view_decorator(refresh_engine_acquire_metrics)
    def acquire_view (self, request: HttpRequest, *args, **kwargs):
        result = self.__acquire_view(request, *args, **kwargs)

        if result is None:
            return JsonResponse({
                "valid": False,
                "token": ""
            }, status=401)
        
        return JsonResponse({
            "valid": True,
            "token": self.encode(result)
        }, status=200)
    def refresh_view (self, request: HttpRequest, *args, **kwargs):
        metrics = refresh_engine_refresh_metrics
        metrics._init()

        start = time.time()
        
        refresh_header = request.headers.get("Refresh")
        if refresh_header is None:
            metrics.total_counter  .run( self.name, time.time() - start )
            metrics.missing_counter.run( self.name, time.time() - start )
            return JsonResponse({
                "valid": False,
                "token": ""
            }, status=400)

        payload = self.decode(refresh_header, verify=False, return_payload=True)
        if payload is None:
            metrics.total_counter.run( self.name, time.time() - start )
            metrics.wrong_counter.run( self.name, time.time() - start )
            return JsonResponse({
                "valid": False,
                "token": ""
            }, status=401)
        if time.time_ns() > payload['rlt']:
            metrics.total_counter.run( self.name, time.time() - start )
            metrics.expired_counter.run( self.name, time.time() - start )
            return JsonResponse({
                "valid": False,
                "token": ""
            }, status=401)

        new_payload = self.__scheme.decode( self.wired_from_payload( payload ) )

        metrics.total_counter.run( self.name, time.time() - start )
        metrics.success_counter.run( self.name, time.time() - start )
        return JsonResponse({
            "valid": True,
            "token": self.encode(new_payload)
        }, status=200)
    
    @property
    def urlpatterns (self):
        return [
            path('acquire/', lambda *args, **kwargs: self.acquire_view(*args, **kwargs)),
            path('refresh/', lambda *args, **kwargs: self.refresh_view(*args, **kwargs))
        ]




from typing import Any

from django.http import HttpRequest, HttpResponse, HttpResponseForbidden

from statelessauth.config import StatelessAuthConfig
from statelessauth.prometheus import RunningOnMetric, middleware_decorator, middleware_metrics

class AuthMiddleware:
    def __init__(self, get_response, config = None):
        self.get_response = get_response

        if config is None:
            config = StatelessAuthConfig()
        self.config = config
    @middleware_decorator
    def __call__(self, request: HttpRequest, *args: Any, **kwds: Any) -> HttpResponse:
        for name, middleware_config in self.config.middlewares:
            with RunningOnMetric(
                middleware_metrics.stateless_auth_middleware_run_per_config,
                middleware_metrics.stateless_auth_middleware_run_per_config_latency,
                [ name ]
            ):
                engine = middleware_config.get_engine()
                field  = middleware_config.field
                
                header, type = middleware_config.header

                header_value = request.headers.get( header, None )

                result = None
                
                if header_value is not None:
                    token = None

                    if type == "" or type == None:
                        token = header_value
                    elif header_value.startswith(type + " "):
                        token = header_value[len(type) + 1:]
                    else:
                        middleware_metrics.stateless_auth_middleware_run_per_config_no_type.labels( name ).inc()

                    if token is not None:
                        result = engine.decode( token, False )

                        if result is None:
                            middleware_metrics.stateless_auth_middleware_run_per_config_wrong_token.labels( name ).inc()
                        else:
                            middleware_metrics.stateless_auth_middleware_run_per_config_valid.labels( name ).inc()

                    if result is None:
                        return HttpResponseForbidden()
                else: 
                    middleware_metrics.stateless_auth_middleware_run_per_config_missing.labels( name ).inc()

                setattr(request, field, result)

        return self.get_response(request, *args, **kwds)




from django.http import HttpRequest
from statelessauth.contrib.auth.models import User

def user_acquire_view (request: HttpRequest) -> "User | None":
    import django.contrib.auth.models as dmodels

    username = request.GET.get("username")
    password = request.GET.get("password")
    if username is None or password is None: return None
    
    users = list( dmodels.User.objects.filter( username = username ) )
    if len(users) == 0: return None
    user = users[0]

    if not user.check_password(password):
        return None

    return User.from_user(user)




from typing import Generic, List, Tuple, TypeVar
from statelessauth.config import StatelessAuthConfig
from statelessauth.prometheus import stateless_auth_decode_decorator, stateless_auth_encode_decorator
from statelessauth.wire import AuthWire

from jose.jws import Key
from jose import jws, JWSError

import json

T = TypeVar("T")

class AuthEngine(Generic[T]):
    # Engine Name
    __name: str = None
    @property
    def name (self):
        return self.__name
    @name.setter
    def name (self, value: str):
        self.__name = value

    __key: "Key | None" = None

    __keyname : str
    __scheme  : AuthWire[T]

    __algorithms : List[str]
    
    def __init__(self, keyname: str, scheme: AuthWire[T], algorithms: List[str] = [ "RS256" ]):
        self.__keyname = keyname
        self.__scheme  = scheme

        assert len(algorithms) != 0
        self.__algorithms = algorithms

    @property
    def key (self) -> Key:
        if self.__key is not None:
            return self.__key
        return StatelessAuthConfig.instance().get_key( self.__keyname )
    @key.setter
    def key (self, value: "Key | None"):
        self.__key = value
    @property
    def urlpatterns (self):
        return []

    def headers (self, data: T):
        return {}
    def payload_from_wired (self, wired):
        return wired
    def wired_from_payload (self, payload):
        return payload
    def validate_payload (self, payload):
        return True

    @stateless_auth_encode_decorator
    def encode (self, data: T):
        wired = self.__scheme.encode(data)
        return jws.sign(self.payload_from_wired(wired), self.key, self.headers(data), self.__algorithms[0])
    @stateless_auth_decode_decorator
    def decode (self, token, verify = True, return_payload = False):
        try:
            payload = jws.verify(token, self.key.public_key(), self.__algorithms)
            payload = json.loads(payload)

            if return_payload:
                return payload

            if not self.validate_payload( payload ):
                raise JWSError()
            
            wired = self.wired_from_payload(payload)

            return self.__scheme.decode(wired)
        except JWSError as error:
            if verify: raise error

            return None



import re
import setuptools

install_requires = open("requirements.txt", "r").readlines()
packages = setuptools.find_packages(
    include=["statelessauth", "statelessauth.engine", "statelessauth.contrib", "statelessauth.contrib.auth"],
    exclude=["statelessauth.test"])

def read(f):
    with open(f, encoding='utf-8') as file:
        return file.read()

def version ():
    return "1.0.1"

setuptools.setup(
    name = "polympiads_statelessauth",
    version = version(),
    description = "Polympiads Framework for Stateless Auth",
    long_description = read("README.md"),
    long_description_content_type='text/markdown',
    url = "https://polympiads.github.io/statelessauth",
    install_requires = install_requires,
    author = "Polympiads",
    license = "MIT",
    packages = packages,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
    project_urls={
        'About Polympiads': 'https://polympiads.ch/',
        'Source': 'https://github.com/polympiads/statelessauth'
    }
)



from typing import Any, Generic, TypeVar

T = TypeVar("T")

class AuthWire(Generic[T]):
    def encode (self, value: T) -> Any:
        raise NotImplementedError()
    def decode (self, value: Any) -> T:
        raise NotImplementedError()




from typing import TYPE_CHECKING, Any, Callable

from statelessauth.contrib.auth.models import Group, Permission, User
from statelessauth.wire import AuthWire

class PermissionWire(AuthWire[Permission]):
    def encode(self, permission: Permission) -> Any:
        return { "name": permission.name, "codename": permission.codename }
    def decode(self, value: Any) -> Permission:
        return Permission(
            value['name'],
            value['codename']
        )

class GroupWire (AuthWire[Group]):
    def __init__(self, permWireClass: Callable[[], AuthWire[Permission]] = PermissionWire):
        self.permWire = permWireClass()
    def encode(self, group: Group) -> Any:
        return {
            "name": group.name,
            "permissions": [
                self.permWire.encode(permission)
                for permission in group.permissions
            ]
        }
    def decode(self, value: Any) -> Group:
        name = value['name']

        encodedPermissions = value['permissions']
        decodedPermissions = [ self.permWire.decode(encoded) for encoded in encodedPermissions ]

        return Group(name, decodedPermissions)

class UserWire (AuthWire[User]):
    def __init__(self, groupWireClass: Callable[[], AuthWire[Group]] = GroupWire):
        self.groupWire = groupWireClass()
    def encode(self, value: User) -> Any:
        return {
            "username": value.username,

            "is_anonymous"     : value.is_anonymous,
            "is_authenticated" : value.is_authenticated,

            "is_staff"     : value.is_staff,
            "is_active"    : value.is_active,
            "is_superuser" : value.is_superuser,

            "groups": [
                self.groupWire.encode(group)
                for group in value.groups
            ]
        }
    def decode(self, value: Any) -> User:
        username = value['username']

        is_anonymous     = value['is_anonymous']
        is_authenticated = value['is_authenticated']

        is_staff     = value['is_staff']
        is_active    = value['is_active']
        is_superuser = value['is_superuser']

        groups = [
            self.groupWire.decode(encodedGroup)
            for encodedGroup in value['groups']
        ]
        return User(username, is_anonymous, is_authenticated, is_staff, is_active, is_superuser, groups)